/* It's highly recommended to use CGL macros instead of changing the current context for plug-ins that perform OpenGL rendering */
#import <OpenGL/CGLMacro.h>

#import "TapControlPlugIn.h"

#define	kQCPlugIn_Name				@"Tap Control"
#define	kQCPlugIn_Description		@"Tap control using an estimator (not very robust).\n\nThis patch uses an estimator that computes the tempo and offset from the taps. But since this estimator does not produce a continuous output, it then uses a PLL to slave an oscillator to the beat, so that it stays continuous yet follows the beat.\n\nFor the correction factor, 0 means the PLL is open and 1 means the PLL corrects 100% of the error. Less than 1 means more stability, but more reaction time, More than 1 means instability."

#define kMaxDelay					3 //After 3 seconds tap sequence is reset
#define kMinDelay					0.15 //Don't take taps closer than 0.15 seconds
#define kMaxNumTaps					12 //Only use the 12 most recent taps

@implementation TapControlPlugIn

/* We need to declare the input / output properties as dynamic as Quartz Composer will handle their implementation */
@dynamic inputTap, inputReset, inputCorrectionFactor;
@dynamic outputEstimatorPhase, outputEstimatorPeriod, outputPLLPhase, outputPLLPeriod;

+ (NSDictionary*) attributes
{
	/* Return the attributes of this plug-in */
	return [NSDictionary dictionaryWithObjectsAndKeys:kQCPlugIn_Name, QCPlugInAttributeNameKey, kQCPlugIn_Description, QCPlugInAttributeDescriptionKey, nil];
}

+ (NSDictionary*) attributesForPropertyPortWithKey:(NSString*)key
{
	/* Return the attributes for the plug-in property ports */
	if([key isEqualToString:@"inputTap"])
	return [NSDictionary dictionaryWithObjectsAndKeys:@"Tap Signal", QCPortAttributeNameKey, nil];
	if([key isEqualToString:@"inputReset"])
	return [NSDictionary dictionaryWithObjectsAndKeys:@"Reset Signal", QCPortAttributeNameKey, nil];
	if([key isEqualToString:@"inputCorrectionFactor"])
	return [NSDictionary dictionaryWithObjectsAndKeys:@"Correction Factor", QCPortAttributeNameKey, [NSNumber numberWithDouble:0.01], QCPortAttributeDefaultValueKey, nil];
	if([key isEqualToString:@"outputEstimatorPhase"])
	return [NSDictionary dictionaryWithObjectsAndKeys:@"Raw Phase", QCPortAttributeNameKey, nil];
	if([key isEqualToString:@"outputEstimatorPeriod"])
	return [NSDictionary dictionaryWithObjectsAndKeys:@"Raw Period", QCPortAttributeNameKey, nil];
	if([key isEqualToString:@"outputPLLPhase"])
	return [NSDictionary dictionaryWithObjectsAndKeys:@"Smoothed Phase", QCPortAttributeNameKey, nil];
	if([key isEqualToString:@"outputPLLPeriod"])
	return [NSDictionary dictionaryWithObjectsAndKeys:@"Smoothed Period", QCPortAttributeNameKey, nil];
	
	return nil;
}

+ (NSArray*) sortedPropertyPortKeys
{
	return [NSArray arrayWithObjects:@"inputTap", @"inputReset", @"inputCorrectionFactor", @"outputPLLPeriod", @"outputPLLPhase", @"outputEstimatorPeriod", @"outputEstimatorPhase", nil];
}

+ (QCPlugInExecutionMode) executionMode
{
	/* This plug-in is a provider (so it always executes) */
	return kQCPlugInExecutionModeProvider;
}

+ (QCPlugInTimeMode) timeMode
{
	/* This plug-in does depend on the time */
	return kQCPlugInTimeModeTimeBase;
}

- (id) init
{
	if(self = [super init])
	_tapTimes = [NSMutableArray new];
	
	return self;
}

- (void) dealloc
{
	[_tapTimes release];
	
	[super dealloc];
}

@end

@implementation TapControlPlugIn (Execution)

- (void) _reset
{
	[_tapTimes removeAllObjects];
	_stillDown = NO;
	
	_estimatorPeriod = 0;
	_estimatorOffset = 0;

	_pllPeriod = 0;
	_pllLastExecuteTime = 0;
	_pllLastExecutePhase = 0;
}

- (BOOL) startExecution:(id<QCPlugInContext>)context
{
	/* Reset out internal state */
	[self _reset];
	
	return YES;
}

- (BOOL) execute:(id<QCPlugInContext>)context atTime:(NSTimeInterval)time withArguments:(NSDictionary*)arguments
{
	/* If we have a reset, reset */
	if(self.inputReset)
		[self _reset];
	
	/* After a few seconds without tapping, reset the tapping sequence - but not the estimator */
	if([_tapTimes lastObject] && ([[_tapTimes lastObject] doubleValue] < (time - kMaxDelay)))
		[_tapTimes removeAllObjects];
	
	/* if we have a continuous tap sequence, only use the last 12 so we can follow the music */
	while([_tapTimes count] >= kMaxNumTaps)
		[_tapTimes removeObjectAtIndex:0];
		
	/* If we have a tap, recalculate the estimator */
	if(self.inputTap) {
		if((!_stillDown)																				// so we don't take the same input twice
		&& ((![_tapTimes lastObject]) || ((time-[[_tapTimes lastObject] doubleValue]) > kMinDelay)))	// so we don't take a possible bounce right after a tap
			{
				/* Record the tap */
				[_tapTimes addObject:[NSNumber numberWithDouble:time]];
				
				/* Update the estimator if we can */
				if([_tapTimes count] >= 2) {
					_estimatorPeriod = ([[_tapTimes lastObject] doubleValue] - [[_tapTimes objectAtIndex:0] doubleValue]) / ([_tapTimes count] - 1);
					if(_estimatorPeriod <= 0)
						_estimatorPeriod = 0.0001;
					
					/* Compute phase offset as mean of distance-to-zero modulo period */
					double phase = 0;
					int i, count = [_tapTimes count];
					for(i=0; i<count; i++) {
						double distanceToZero = fmod([[_tapTimes objectAtIndex:i] doubleValue] / _estimatorPeriod, 1);
						if(distanceToZero > 0.5)
							distanceToZero -= 1;
						phase += distanceToZero;
					}
					_estimatorOffset  = phase / count;
				}
			}
		_stillDown = YES;
	}
	else
		_stillDown = NO;
	
	/* Now update our output oscillator so it synchronizes slowly to the estimator */
	
	/* Initialize the oscillator if needed */
	if(_pllLastExecuteTime <= 0)
		_pllLastExecuteTime = time;

	/* Perform the actual oscillator calculation */
	double pllPhase = 0;
	if((_pllPeriod > 0) && (_pllLastExecuteTime < time))
		pllPhase = _pllLastExecutePhase + (time - _pllLastExecuteTime)/_pllPeriod;					
	pllPhase = fmod(pllPhase, 1);
	_pllLastExecutePhase = pllPhase;
	_pllLastExecuteTime = time;
	self.outputPLLPhase = pllPhase;
	self.outputPLLPeriod = _pllPeriod;
	
	/* Perform the period correction to slave the oscillator to the estimator */
	double estimatorPhase = 0;
	if(_estimatorPeriod > 0) {
		estimatorPhase = fmod( time/_estimatorPeriod + _estimatorOffset, 1);
		double phaseError = estimatorPhase - pllPhase; // positive is pll is lagging
		if(phaseError > 0.5)
			phaseError = phaseError - 1;
		if(phaseError < -0.5)
			phaseError = phaseError + 1;
		_pllPeriod = _estimatorPeriod + phaseError * self.inputCorrectionFactor;
	}
	self.outputEstimatorPhase = estimatorPhase;
	self.outputEstimatorPeriod = _estimatorPeriod;
	
	return YES;
}

@end
